#define G_LOG_DOMAIN "boilerplate-window"

#include "config.h"

#include "boilerplate-window.h"
#include "boilerplate-application.h"

#include <gtk/gtk.h>

/**
 * SECTION:boilerplate-window
 * @title: BoilerplateWindow
 * @short_description: A #GtkApplicationWindow instance created for
 * #BoilerplateApplication
 * @include: boilerplate-window.h
 * @see_also: #GtkApplicationWindow
 *
 * The #BoilerplateApplication is an instance #GtkApplicationWindow, and acts as
 * the primary window for the current #GtkApplication.
 **/

typedef struct
{
    gint width;
    gint height;
    gint x;
    gint y;
    gboolean maximized;
    gboolean fullscreen;
} BoilerplateWindowState;

typedef struct _BoilerplateWindowPrivate
{
    GtkWidget *toolbar;
    GtkWidget *headerbar;

    GtkWidget *content;

    guint timeout_id;
} BoilerplateWindowPrivate;

struct _BoilerplateWindow
{
    GtkApplicationWindow parent_instance;

    BoilerplateWindowState state;
};

#define STATE_TIMEOUT 100 /* ms */

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateWindow, boilerplate_window, GTK_TYPE_APPLICATION_WINDOW)

/*
 * on_close:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * An action wrapper to close the current window.
 */
static void
on_close (GSimpleAction *action,
          GVariant      *state,
          gpointer       user_data)
{
    g_debug ("Requesting window be closed");
    gtk_window_close (GTK_WINDOW (user_data));
}

/*
 * on_fullscreen:
 * @action: A pointer to the object that originated the callback. In this
 * instance a #GSimpleAction, which will be added to the #GAction menu
 * @parameter: arg_data from the #GActionEntry
 * @user_data: User data set when the signal handler was connected, in this
 * instance it should be a #GtkApplication
 *
 * An action wrapper to toggle the window to fullscreen.
 */
static void
on_fullscreen (GSimpleAction *action,
               GVariant      *parameter,
               gpointer       user_data)
{
    boilerplate_window_toggle_fullscreen (BOILERPLATE_WINDOW (user_data));
}

const GActionEntry window_entries[] =
{
    { "close", on_close },
    { "fullscreen", on_fullscreen },
};

/*
 * boilerplate_window_set_fullscreen:
 * @self: An instance of #BoilerplateWindow
 *
 * Sets the @self instance to fullscreen mode.
 */
void
boilerplate_window_toggle_fullscreen (BoilerplateWindow *self)
{
    BOILERPLATE_IS_WINDOW (self);

    GdkWindow *window = gtk_widget_get_window (GTK_WIDGET (self));
    GdkWindowState state = gdk_window_get_state (window);

    if (state & GDK_WINDOW_STATE_FULLSCREEN) {
        gtk_window_unfullscreen (GTK_WINDOW (self));
        g_debug ("Window is requesting to unfullscreen");
    } else {
        gtk_window_fullscreen (GTK_WINDOW (self));
        g_debug ("Window is requesting to fullscreen");
    }

    /*
     * DEBUG CODE:
     * Check new state, after request
     */
    state = gdk_window_get_state (window);

    if (state & GDK_WINDOW_STATE_FULLSCREEN) {
        g_debug ("Window is unfullscreen");
    } else {
        g_debug ("Window is fullscreen");
    }
}

/*
 * boilerplate_window_state_store:
 * @self: An instance of #BoilerplateWindow
 *
 * Saves the `width`, `height`, `x`, `y`, `maximized` and `fullscreen` values
 * from from the #BoilerplateWindowState struct into the gsettings:
 * PACKAGE_ID ".window-state", eg. org.gnome.app.window-state
 */
static void
boilerplate_window_state_store (BoilerplateWindow *self)
{
    GSettings *settings = g_settings_new (PACKAGE_ID ".window-state");
    gboolean status;

    status = g_settings_set_int (settings, "width", self->state.width);

    if (status)
        g_debug ("WindowSate: Saved width      = %i", self->state.width);
    else
        g_debug ("WindowState: Could not save width");


    status = g_settings_set_int (settings, "height", self->state.height);

    if (status)
        g_debug ("WindowSate: Saved height     = %i", self->state.height);
    else
        g_debug ("WindowState: Could not save height");

    status = g_settings_set_int (settings, "x", self->state.x);

    if (status)
        g_debug ("WindowSate: Saved position-x = %i", self->state.height);
    else
        g_debug ("WindowState: Could not save postion-x");

    status = g_settings_set_int (settings, "y", self->state.y);

    if (status)
        g_debug ("WindowSate: Saved position-y = %i", self->state.height);
    else
        g_debug ("WindowState: Could not save positon-y");

    status = g_settings_set_boolean (settings, "maximized", self->state.maximized);

    if (status)
        g_debug ("WindowSate: Saved maximized  = %s", (self->state.maximized) ? "TRUE" : "FALSE");
    else
        g_debug ("WindowState: Could not save maximized");

    status = g_settings_set_boolean (settings, "fullscreen", self->state.fullscreen);

    if (status)
        g_debug ("WindowSate: Saved fullscreen = %s", (self->state.fullscreen) ? "TRUE" : "FALSE");
    else
        g_debug ("WindowState: Could not save fullscreen");
}

/*
 * boilerplate_window_state_load:
 * @self: An instance of #BoilerplateWindow
 *
 * Loads the `width`, `height`, `x`, `y`, `maximized` and `fullscreen` values
 * from gsettings into the #BoilerplateWindowState struct, located in the
 * #BoilerplateWindow struct.
 */
static void
boilerplate_window_state_load (BoilerplateWindow *self)
{
    GSettings *settings = g_settings_new (PACKAGE_ID ".window-state");

    self->state.width = g_settings_get_int (settings, "width");
    g_debug ("WindowSate: Loaded width      = %i", self->state.width);

    self->state.height = g_settings_get_int (settings, "height");
    g_debug ("WindowSate: Loaded height     = %i", self->state.height);

    self->state.x = g_settings_get_int (settings, "x");
    g_debug ("WindowSate: Loaded position-x = %i", self->state.x);

    self->state.y = g_settings_get_int (settings, "y");
    g_debug ("WindowSate: Loaded position-y = %i", self->state.y);

    self->state.maximized = g_settings_get_boolean (settings, "maximized");
    g_debug ("WindowSate: Loaded maximized  = %s", (self->state.maximized) ? "TRUE" : "FALSE");

    self->state.fullscreen = g_settings_get_boolean (settings, "fullscreen");
    g_debug ("WindowSate: Loaded fullscreen = %s", (self->state.fullscreen) ? "TRUE" : "FALSE");
}

/*
 * boilerplate_window_state_set:
 * @user_data: Data passed into the function via g_timeout_add(), in this case
 * an instance of #BoilerplateWindow
 *
 * Saves the information from the current window into the WindowState. Is called
 * from boilerplate_window_configure_event() and boilerplate_window_window_state_event().
 *
 * Returns: #gboolean; Sets %FALSE when the function does what it was going todo
 * (save window state) and set timeout in to 0.
 */
static gboolean
boilerplate_window_state_set (gpointer user_data)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (user_data);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    GdkWindow *window = gtk_widget_get_window (GTK_WIDGET (self));
    GdkWindowState state = gdk_window_get_state (window);

    self->state.maximized  = (state & GDK_WINDOW_STATE_MAXIMIZED) ? 1 : 0;
    self->state.fullscreen = (state & GDK_WINDOW_STATE_FULLSCREEN) ? 1 : 0;

    if (self->state.maximized || self->state.fullscreen)
    {
        priv->timeout_id = 0;
        return FALSE;
    }

    gtk_window_get_size (GTK_WINDOW (self),
                         &self->state.width,
                         &self->state.height);

    //TODO: Make saving postion x and y work in Wayland?
    gtk_window_get_position (GTK_WINDOW (self),
                             &self->state.x,
                             &self->state.y);

    priv->timeout_id = 0;

    return FALSE;
}

/*
 * boilerplate_window_configure_event:
 * @widget: An instance of #BoilerplateWindow
 * @event: The current state of the window as a #GdkEventWindowState
 *
 * Saves the window state/geometry when it changes. Calls
 * boilerplate_window_state_set().
 *
 * Returns: The #gboolean return value (retval); %TRUE the event happened or
 * %FALSE the event did not happen
 */
static gboolean
boilerplate_window_configure_event (GtkWidget         *widget,
                                    GdkEventConfigure *event)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);
    gboolean retval;

    if (priv->timeout_id != 0)
    {
        g_source_remove (priv->timeout_id);
        priv->timeout_id = 0;
    }

    priv->timeout_id = g_timeout_add (STATE_TIMEOUT,
                                      boilerplate_window_state_set,
                                      self);

    retval = GTK_WIDGET_CLASS (boilerplate_window_parent_class)->configure_event (widget, event);

    return retval;
}

/*
 * boilerplate_window_destroy:
 * @widget: An instance of #BoilerplateWindow
 * @event: The current state of the window as a #GdkEventWindowState
 *
 * Saves the window state/geometry when it changes, a.k.a if the window is
 * maximized or fullscreen. Calls boilerplate_window_state_set().
 *
 * Returns: The #gboolean return value (retval); %TRUE the event happened or
 * %FALSE the event did not happen
 */
static gboolean
boilerplate_window_window_state_event (GtkWidget           *widget,
                                       GdkEventWindowState *event)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);
    gboolean retval;

    if (priv->timeout_id != 0)
    {
        g_source_remove (priv->timeout_id);
        priv->timeout_id = 0;
    }

    priv->timeout_id = g_timeout_add (STATE_TIMEOUT,
                                      boilerplate_window_state_set,
                                      self);

    retval = GTK_WIDGET_CLASS (boilerplate_window_parent_class)->window_state_event (widget, event);

    return retval;
}

/*
 * boilerplate_window_constructed:
 * @object: An instance of #BoilerplateWindow
 *
 * The constructed virtual function sets first-run settings for the
 * #BoilerplateWindow object.
 *
 * It sets:
 *  * Window dimensions
 *  * Window postion
 *  * Window maximized
 *  * Window fullscreen
 */
static void
boilerplate_window_constructed (GObject *object)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);

    self->state.width      = -1;
    self->state.height     = -1;
    self->state.x          =  0;
    self->state.y          =  0;
    self->state.maximized  =  FALSE;
    self->state.fullscreen =  FALSE;

    boilerplate_window_state_load (self);

    gtk_window_set_default_size (GTK_WINDOW (self),
                                 self->state.width,
                                 self->state.height);

    gtk_window_set_position (GTK_WINDOW (self), GTK_WIN_POS_MOUSE);

    if (self->state.x && self->state.y)
        gtk_window_move (GTK_WINDOW (self),
                         self->state.x,
                         self->state.y);

    if (self->state.maximized)
        gtk_window_maximize (GTK_WINDOW (self));

    if (self->state.fullscreen)
        gtk_window_fullscreen (GTK_WINDOW (self));

    G_OBJECT_CLASS (boilerplate_window_parent_class)->constructed (object);
}

/*
 * boilerplate_window_finalize:
 * @object: An instance of #BoilerplateWindow
 *
 * Does any last minute clean-ups when the class ends.
 */
static void
boilerplate_window_finalize (GObject *object)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);

    boilerplate_window_state_store (self);

    G_OBJECT_CLASS (boilerplate_window_parent_class)->finalize(object);
}

/*
 * boilerplate_window_init:
 * @self: An instance of #BoilerplateWindow
 *
 * Initializes various aspects of the window. Anything the window would need
 * before becoming presented should be done here.
 */
static void
boilerplate_window_init (BoilerplateWindow *self)
{
    GApplication *application = g_application_get_default ();
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    const gchar *fullscreen_accels[] =
    {
        "F11",
        "<Primary>f",
        NULL
    };

    /* Init window.ui template */
    gtk_widget_init_template (GTK_WIDGET (self));

    /* Set window title */
    gtk_window_set_title (GTK_WINDOW (self), PACKAGE_NAME);

    /* Set default window decorations (a.k.a GtkHeaderBar) */
    gtk_window_set_titlebar (GTK_WINDOW (self), priv->headerbar);
    gtk_container_add (GTK_CONTAINER (priv->headerbar), priv->toolbar);

    /* Add keybind */
    g_action_map_add_action_entries (G_ACTION_MAP (self),
                                     window_entries, G_N_ELEMENTS (window_entries),
                                     self);

    boilerplate_application_set_accelerator (GTK_APPLICATION (application), "win.close", "<Primary>w");
    boilerplate_application_set_accelerators (GTK_APPLICATION (application), "win.fullscreen", fullscreen_accels);
}

/*
 * boilerplate_window_class_init:
 * @klass: An instance of #BoilerplateWindowClass
 *
 * Initializes the various virtural functions in #BoilerplateWindowClass. In
 * addtition it connects the various #GtkWidgets from the gresource file using
 * gtk_widget_class_set_template_from_resource().
 */
static void
boilerplate_window_class_init (BoilerplateWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

    object_class->constructed = boilerplate_window_constructed;
    object_class->finalize = boilerplate_window_finalize;

    widget_class->configure_event = boilerplate_window_configure_event;
    widget_class->window_state_event = boilerplate_window_window_state_event;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 PACKAGE_RESOURCES "/ui/window.ui");

    gtk_widget_class_bind_template_child_private (widget_class, BoilerplateWindow, toolbar);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplateWindow, headerbar);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplateWindow, content);
}

/**
 * boilerplate_window_new:
 * @application: A #GtkApplication instance
 *
 * Creates a new #BoilerplateWindow, a sub-object of #GtkApplicationWindow.
 *
 * ## Usage
 * |[
 * BoilerplateWindow *window;
 *
 * window = boilerplate_window_new (application);
 * ]|
 *
 * When creating a boilerplate_window_new() instance it's recommended to set the
 * instance into a variable inside the #BoilerplateApplication struct. The
 * window could also be added to a #GList, so multiple windows can belong to
 * the application.
 *
 * Returns: #BoilerplateWindow; a sub-object of #GtkApplicationWindow
 */
BoilerplateWindow *
boilerplate_window_new (BoilerplateApplication *application)
{
    return g_object_new (BOILERPLATE_TYPE_WINDOW,
                         "application", application,
                         "title", PACKAGE_NAME,
                         NULL);
}

